"""
Lösungen für die Übungen zur OSM-Datenabfrage und Visualisierung
=============================================================

Dieses Skript enthält die Lösungen für die Übungen im Arbeitsblatt.
Es ist in mehrere Abschnitte unterteilt:
1. OSM-Datenabfrage und Visualisierung
2. Geokodierung von Adressen
"""

# =============================================================================
# 1. OSM-Datenabfrage und Visualisierung
# =============================================================================

# Import der benötigten Bibliotheken
import plotly.express as px  # Für interaktive Karten
import geopandas as gpd     # Für die Arbeit mit Geodaten
import matplotlib.pyplot as plt  # Für statische Karten
from collections import namedtuple  # Für strukturierte Daten
from osm2geojson import overpass_call, json2geojson  # Für OSM-Datenabfragen

# Definition von Hilfsstrukturen für die Datenorganisation
Bbox = namedtuple("Bbox", ["south", "west", "north", "east"])  # Bounding Box für geografische Gebiete
Tag = namedtuple("Tag", ["key", "values"])  # OSM-Tags für die Suche

def load_osm_from_overpass(bbox, tag, crs="epsg:4326") -> gpd.GeoDataFrame:
    """
    Lädt OSM-Daten von der Overpass API und konvertiert sie in ein GeoDataFrame.
    
    Args:
        bbox: Bounding Box für das Suchgebiet
        tag: OSM-Tag für die Suche (z.B. sport=table_tennis)
        crs: Koordinatenreferenzsystem (Standard: WGS84)
    
    Returns:
        GeoDataFrame mit den gefundenen OSM-Elementen
    """
    geojson = load_osm_from_overpass_geojson(bbox, tag, crs)
    return gpd.GeoDataFrame.from_features(geojson, crs=crs)

def load_osm_from_overpass_geojson(bbox, tag, crs="epsg:4326"):
    """
    Erstellt und führt eine Overpass API-Abfrage durch.
    
    Args:
        bbox: Bounding Box für das Suchgebiet
        tag: OSM-Tag für die Suche
        crs: Koordinatenreferenzsystem
    
    Returns:
        GeoJSON-Objekt mit den gefundenen OSM-Elementen
    """
    values = None if "*" in tag.values else tag.values
    query = f"""
    [out:json];
    node["{tag.key}"="{values[0] if values else '*'}"]({bbox.south},{bbox.west},{bbox.north},{bbox.east});
    out body;
    >;
    out skel qt;
    """
    response = overpass_call(query)
    return json2geojson(response)

def plot_gdf_on_map(gdf, color="k", title=""):
    """
    Erstellt eine interaktive Karte mit den GeoDataFrame-Daten.
    
    Args:
        gdf: GeoDataFrame mit den zu visualisierenden Daten
        color: Farbe für die Darstellung
        title: Titel der Karte
    
    Returns:
        Plotly Figure-Objekt mit der interaktiven Karte
    """
    fig = px.scatter_mapbox(
        gdf,
        lat=gdf.geometry.y,
        lon=gdf.geometry.x,
        color_discrete_sequence=["fuchsia"],
        height=500,
        zoom=11,
    )
    fig.update_layout(title=title, mapbox_style="open-street-map")
    fig.update_layout(margin={"r": 0, "l": 0, "b": 0})
    return fig

def main():
    """
    Hauptfunktion zum Testen der OSM-Datenabfrage.
    Führt zwei Tests durch:
    1. Tischtennistische in Zürich
    2. Trinkbrunnen in Rapperswil-Jona
    """
    # Test 1: Tischtennistische in Zürich
    print("Test 1: Tischtennistische in Zürich")
    tag = Tag(key="sport", values=["table_tennis"])  # OSM-Tag für Tischtennistische
    bbox_zurich = Bbox(west=8.471668, south=47.348834, east=8.600454, north=47.434379)
    table_tennis_gdf = load_osm_from_overpass(bbox_zurich, tag)
    print(f"Gefundene Tischtennistische: {len(table_tennis_gdf)}")
    
    # Test 2: Trinkbrunnen in Rapperswil-Jona
    print("\nTest 2: Trinkbrunnen in Rapperswil-Jona")
    tag = Tag(key="amenity", values=["drinking_water"])  # OSM-Tag für Trinkbrunnen
    bbox_rapperswil = Bbox(west=8.8000, south=47.2000, east=8.9000, north=47.3000)
    drinking_water_gdf = load_osm_from_overpass(bbox_rapperswil, tag)
    print(f"Gefundene Trinkbrunnen: {len(drinking_water_gdf)}")

# =============================================================================
# 2. Geokodierung von Adressen
# =============================================================================

def geocode_address(address):
    """
    Geokodiert eine Adresse mit der Nominatim API.
    Konvertiert eine textuelle Adresse in geografische Koordinaten.
    
    Args:
        address: Textuelle Adresse (z.B. "Vulkanstrasse 106, Zürich")
    
    Returns:
        WKT-String mit den Koordinaten im Format 'SRID=4326;POINT(lon lat)'
        oder None, wenn keine Koordinaten gefunden wurden
    
    Hinweis: Bitte beachten Sie die Nutzungsrichtlinien der Nominatim API:
    https://operations.osmfoundation.org/policies/nominatim/
    """
    import requests
    from time import sleep
    
    base_url = 'https://nominatim.openstreetmap.org/search?'
    params = {
        'q': address,
        'format': 'json',
        'limit': 1,
        'User-Agent': 'OpenSchoolMaps Tutorial'  # Wichtig: Einen spezifischen User-Agent verwenden
    }
    
    try:
        response = requests.get(base_url, params=params)
        response.raise_for_status()  # Prüft auf HTTP-Fehler
        loc = response.json()
        
        if len(loc) == 0:
            return None
        
        # Koordinaten im WKT-Format zurückgeben
        wkt = f'SRID=4326;POINT({loc[0]["lon"]} {loc[0]["lat"]})'
        sleep(1)  # Throttling gemäß Nominatim-Richtlinien
        return wkt
        
    except requests.exceptions.RequestException as e:
        print(f"Fehler bei der Geokodierung: {e}")
        return None

def test_geocoding():
    """
    Testet die Geokodierungsfunktion mit der Beispieladresse.
    Zeigt die gefundenen Koordinaten an oder eine Fehlermeldung.
    """
    address = 'Vulkanstrasse 106, Zürich'
    result = geocode_address(address)
    
    if result:
        print(f"Geokodierte Koordinaten für '{address}':")
        print(result)
    else:
        print(f"Keine Koordinaten gefunden für '{address}'")

# =============================================================================
# Hauptprogramm
# =============================================================================

if __name__ == "__main__":
    print("=== Start der OSM-Datenabfragen ===")
    main()  # Führt die OSM-Datenabfragen durch
    
    print("\n=== Test der Geokodierungsfunktion ===")
    test_geocoding()  # Testet die Geokodierungsfunktion
